Attribute VB_Name = "Alpha"
Option Explicit
Private Const OneSecond = 1 / 24 / 60 / 60

Public Sub Main()

    ' Called when program loads.
    Load frmComm
    frmComm.Show
       
       
    ' The AvBrand AlphaSign communications module supports two methods of communication:
    ' Serial port - for a sign directly connected to the computer.
    ' TCP - for a sign connected via a Ethernet to Serial adapter, such as the one sold by Sparkfun for $30.
    
    ' Each "sign" object represents one sign, or all signs.
    ' In this example code, the code has been configured for two signs, with IDs 01 and 02.
    ' However, we provide both signs with the same initial programming using ID 00.
    
    ' Here you need to provide the com port object, the com port number, and a picturebox object.
    ' The picturebox is used for if you are loading graphics in.
    
    Set frmComm.MySign1 = New AlphaSignComms
    frmComm.MySign1.SetObjects frmComm, 5
    frmComm.MySign1.setSignID "00"
    
    Set frmComm.MySign2 = New AlphaSignComms
    frmComm.MySign2.setUseOtherObject frmComm.MySign1
    frmComm.MySign2.setSignID "02"
    
    DoEvents
    
    
    ' SIGN PAGES IN USE
    ' A = Sample page
    ' B = Sample page 2
    ' Q = Graphic 1
    ' 1 = String 1
    ' 2 = String 2
    
    ' We're going to work with the "All signs" object.
    With frmComm.MySign1
        ' Tell the class we're starting a new programming session.
        .BeginNew
        
        ' Configure the strings we will be using and the amount of memory we require.
        .MakeString "1", 125    ' 125 bytes of memory. This is the max for strings.
        .MakeString "2", 125    ' 125 bytes of memory
        
        ' Load the sample image into file Q.
        .loadPictureIntoMemory "Q", App.Path & "\Sample Graphics\signal.bmp", ALPHA_Monochrome
                
                
        ' We're going to begin a new PAGE in the sign.
        ' File A: Sample 1
        .newTextFile
        
        ' This is for two-line signs... this next message will spray onto the top line.
        .setCurrentLine lineTop
        .addText "<SPRAY>Hello, World!<NOHOLD>"
        
        ' We're using a <NOHOLD> tag to indicate we don't want the sign to pause between rendering the top line and the next line.
        
        .setCurrentLine lineBottom
        .addText "<ROLL:UP>I am just fine!<WIPE:IN>"
        
        ' The transition specified at the END of the text indicates how we want the text to disappear.
        .setCurrentLine lineMiddle
        
        ' Display our sample picture
        .addText "<SCROLL><PIC>Q"
        
        
        ' Display some animations.
        .addText "<ANIM:DRINKDRIVE><ANIM:SLOTS><ANIM:THANKYOU>"
        
        
        ' This adds the PAGE to the memory and provides the name of the file.
        .finishTextFile "A"
        
            
            
        ' Create the second sample file.
        .newTextFile
        .setCurrentLine lineTop
        .addText "<HOLD>" ' the <HOLD> transition just makes items appear without animating.

        ' Fixed Width makes each character in the text the same width. Not all signs support this.
        ' These items display the contents of Strings 1 and 2 on two lines.
        .addText "<FIXEDWIDTH:ON>Item 1: <STRING>1<NOHOLD>"

        .setCurrentLine lineBottom
        .addText "<HOLD>"
        .addText "<FIXEDWIDTH:ON>Item 2: <STRING>2"

        .finishTextFile "B"
        
        
        
        '.newTextFile ' For some reason, the sign loses the last file.  So this is here for the sign to eat so it doesn't eat our file above.
        '.addText "Junk"
        '.finishTextFile "@"
                
        
        
        ' Now that we've configured the graphics, programming, and strings, we need to send this data to the sign.
        .SendToSign False
            
        ' The RUN SEQUENCE tells the sign which pages we want to display. Let's start with just page 'A'.
        .setRunSequence "A"
        
        
        ' Set the initial values into the strings.
        .setString "1", "Initial"
        .setString "2", "Value"
        
        
    End With
    

End Sub



