VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "AlphaSignComms"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' ****************************************************************************************************
' ****************************************************************************************************
' ****************************************************************************************************
'     AVBRAND.COM presents:
'       ALPHA (AMS) SIGN INTERFACE LIBRARY
'
'     This class makes it super-easy to communicate with an 'Alpha' or Adaptive Micro Systems' brand LED sign.
'     See www.avbrand.com
'
'     VERSION 1.1
'     - Added 'SignSetting' property so you can store custom settings on each Sign object.
'     - Added some limited support for COM buffering, this isn't really needed.
'     - Fixed bugs with adding pictures
'     - Added errors if you try to create a string longer than 125 bytes
'     - If you try to set a string longer than 125, it crops it.
'     - Added 'SendBeep' method to make a sign beep on command
'     - Added a few methods that read data from the signs, such as the error register, and 'check if sign is alive'.
'     - Support for checksums, but these aren't fully implemented
'     - The form no longer needs to have the MSComm object on it, but it does need to be referenced, and 'Remove information about unused Active-X Controls' needs to be switched off.
'
'     VERSION 1.0
'      Initial Release
'     RELEASED JULY 13 2010 UNDER THE GPL LICENSE
'
' ****************************************************************************************************
' ****************************************************************************************************
' ****************************************************************************************************


Option Explicit
Private Declare Sub Sleep Lib "kernel32" (ByVal dwMilliseconds As Long)

Public Event TcpConnectionActive()
Public Event Progress(nOf100 As Long)


Public Enum lineOptions
    lineMiddle = 0
    lineTop = 1
    lineBottom = 2
    lineFill = 3
End Enum
Public Enum ALPHA_ColorOptions
    ALPHA_Monochrome = 1
    ALPHA_3Color = 2
    ALPHA_8Color = 3
End Enum

Public Enum ALPHA_FileTypes
    ALPHA_TEXT = 1
    ALPHA_STRING = 2
    ALPHA_DOTS = 3
End Enum

Private Enum eCommModes
    eSerial
    eTCP
    eUDP
End Enum

Private WithEvents COMM_OBJ As MSComm
Attribute COMM_OBJ.VB_VarHelpID = -1
Private WithEvents TCP_OBJ As MSWinsockLib.Winsock
Attribute TCP_OBJ.VB_VarHelpID = -1
Private WithEvents MyTimer As VB.Timer
Attribute MyTimer.VB_VarHelpID = -1
Private picBox As PictureBox

Private COM_PORT As Long
Private TCP_PORT As Long
Private TCP_HOST As String
Private COMM_MODE As eCommModes
Private COMMREADY As Boolean


Private usingOtherSerial As Boolean
Private otherSerialObj As AlphaSignComms

Private pendingData() As String
Private pendingMemory As String

Private buildText As String


Private receiveBuffer As String
Private currentLine As String

Private mySignID As String

Private MAX_WORD_LENGTH As Long

' PUBLIC MEMBERS
Public LastRunSequence As String

' EVENTS
Public Event CommError(errMsg As String)


Private TCP_BUFFER As String
Private firstTCP As Boolean

Private lastConnAttempt As Date

Private Const OneSecond = 1 / 24 / 60 / 60

Private outputBuffer As String

Public U_SignID As String
Private U_SignSettings As New Dictionary

Public U_BlockUpdatesUntil As Date
'Public U_ResetRunSequenceAt As Date
Public U_LastSignOptUpdate As Date
Public U_NextRSS As Date

Public U_NextPageChangeAt As Double
Public U_LastDataUpdate As Date


Public LastLiveCheck As Date
Public LastLiveResponse As Date


Dim COMM_PACKET As String
Dim COMM_BUFFER As New Collection
Dim USE_COMM_BUFFER As Boolean
Dim theForm As Form


Dim IsTransmitting As Boolean





Private Sub Class_Initialize()
    ' Configure our startup parameters.
    mySignID = "00"
    
    ' The maximum length of a word on the sign. Increase this number if you have a longer sign.
    ' This is necessary because the Alpha sign has issues with long words, it causes it to crash (since it can't wrap them properly).
    MAX_WORD_LENGTH = 19
End Sub

Public Property Let SignSetting(vName, vValue)
    If VarType(vValue) = vbObject Then
        Set U_SignSettings(vName) = vValue
    Else
        U_SignSettings(vName) = vValue
    End If
End Property

Public Property Get SignSetting(vName)
    If VarType(U_SignSettings(vName)) = vbObject Then
        Set SignSetting = U_SignSettings(vName)
    Else
        SignSetting = U_SignSettings(vName)
    End If
End Property

Public Function DoesSettingExist(vName) As Boolean
    DoesSettingExist = U_SignSettings.Exists(vName)
End Function

Public Sub SetObjects(InForm As Form, COMPORT As Long, Optional UseCommBuffer As Boolean)
    ' Provide the objects you want me to use.
    usingOtherSerial = False
    Set theForm = InForm
    
    ' Create a com port for us.
    Set COMM_OBJ = InForm.Controls.Add("MSCommLib.MSComm", "MyComAlpha" & Int(Timer) & "_" & Int(Rnd * 10000))
    
    ' Create a picture box too.
    Set picBox = InForm.Controls.Add("VB.PictureBox", "MyPicAlpha" & Int(Timer) & "_" & Int(Rnd * 10000))
    
    COM_PORT = COMPORT
    COMMREADY = False
    
    USE_COMM_BUFFER = UseCommBuffer
        
    If USE_COMM_BUFFER Then
        Set MyTimer = InForm.Controls.Add("VB.Timer", "MyTimerAlpha" & Int(Timer) & "_" & Int(Rnd * 10000))
        MyTimer.Interval = 1
        MyTimer.Enabled = False
    End If
    
        
    COMM_MODE = eSerial
    openComm
End Sub


Private Sub Class_Terminate()
    On Error Resume Next
    If Not MyTimer Is Nothing Then
        theForm.Controls.Remove MyTimer
    End If
    If Not picBox Is Nothing Then
        theForm.Controls.Remove picBox
    End If
    If Not COMM_OBJ Is Nothing Then
        theForm.Controls.Remove COMM_OBJ
    End If
    If Not TCP_OBJ Is Nothing Then
        theForm.Controls.Remove TCP_OBJ
    End If

End Sub


Public Sub SetObjectsTCP(InForm As Form, Host As String, TCPPORT As Long)
    ' For use with TCP objects.
    
    Set TCP_OBJ = InForm.Controls.Add("MSWinsock.winsock.1", "MyWinSockAlpha" & Int(Timer) & "_" & Int(Rnd * 10000))

    ' Create a picture box too.
    Set picBox = InForm.Controls.Add("VB.PictureBox", "MyPicAlpha" & Int(Timer) & "_" & Int(Rnd * 10000))

    usingOtherSerial = False
    TCP_PORT = TCPPORT
    TCP_HOST = Host
    COMMREADY = False
    
    COMM_MODE = eTCP
    openComm
End Sub

Public Sub setUseOtherObject(otherASC As AlphaSignComms)
    ' We don't have our own serial port, instead, we are using one attached to another object.
    usingOtherSerial = True
    Set otherSerialObj = otherASC
End Sub

Private Sub openComm()
    On Error Resume Next
    ' Open up the comm port.
    
    Select Case COMM_MODE
        Case eSerial
            If COMM_OBJ.PortOpen Then COMM_OBJ.PortOpen = False
            With COMM_OBJ
                .CommPort = COM_PORT
                .Settings = "9600,n,8,1"
                .SThreshold = 1
                .RThreshold = 1
                .InputMode = comInputModeText
                .OutBufferSize = 512
                .PortOpen = True
                COMMREADY = True
            End With
            
            If Err.Number <> 0 Then
                Debug.Print Err.Description
                
                RaiseEvent CommError(Err.Description & vbCrLf & "Port " & COM_PORT)
            End If
        Case eTCP
            TCP_OBJ.Close
            
            lastConnAttempt = Now
            
            Debug.Print "!!! Connecting to sign...", TCP_HOST, TCP_PORT
            With TCP_OBJ
                .Connect TCP_HOST, TCP_PORT
            End With
            
            If Err.Number <> 0 Then
                RaiseEvent CommError(Err.Description)
            End If
        
    End Select
End Sub


Public Function setSignID(ID As String)
    ' Provide the ID that is my sign.
    mySignID = ID
End Function

Private Function Receive(inDat As String)
    ' Command received from sign.
    
    receiveBuffer = receiveBuffer & inDat
    'Debug.Print "Adding", inDat, " to buffer ", Len(receiveBuffer)
    
    ' Do we have an end?
    Dim e As Long, f As Long
    Dim Msg As String
    
    'Debug.Print "[RECV]", inDat
    
    If safeRight(receiveBuffer, 4) = "next" Then
        IsTransmitting = False
    End If
    
    e = InStr(receiveBuffer, Chr(1))
    f = InStr(e + 1, receiveBuffer, Chr(4))
    ' Extract
    If e > 0 And f > e Then
            
        Msg = Mid(receiveBuffer, e + 1, f - e - 1)
        receiveBuffer = Right(receiveBuffer, Len(receiveBuffer) - f)
        
        processMessage Msg
        
    End If
    
    
End Function

Public Function BeginNew()
    ' Start a new session.
    pendingMemory = ""
    ReDim pendingData(0)
    
    Debug.Print "New session begun."
End Function

Public Function newTextFile()
    ' Start a new text file.
    
    buildText = ""
    setCurrentLine lineMiddle

End Function

Public Function setCurrentLine(ln As lineOptions)
    ' Set the current line on which text appears.
    ' Need at least a 2-line sign to use this.
    
    Select Case ln
        Case lineOptions.lineMiddle: currentLine = " "
        Case lineOptions.lineTop: currentLine = """"
        Case lineOptions.lineBottom: currentLine = "&"
        Case lineOptions.lineFill: currentLine = "0"
    End Select

End Function


Public Function addText(Msg As String)
    ' Add to the text file we are building.
    buildText = buildText & makeAlpha(Msg)
End Function

Public Function finishTextFile(fileLabel As String, Optional noAllocate As Boolean, Optional neverPlay As Boolean)
    ' Close the text file by allocating the memory and adding the data to the transfer queue.
    
    Dim s As String
    Dim nMem As Long
    nMem = Len(buildText) + 8 ' Allocate a bit extra just in case
    
    ' Allocate memory.
    If Not noAllocate Then
        Debug.Print "Finishing text file ", fileLabel, " by allocating ", nMem, " bytes of memory."
        If neverPlay Then
            allocateMemory fileLabel, ALPHA_TEXT, nMem, True, , , , "FE", "00"
        Else
            allocateMemory fileLabel, ALPHA_TEXT, nMem, True
        End If
    End If
    
    s = ""
    s = s & "A"         ' Write TEXT file
    s = s & fileLabel   ' Label
    s = s & buildText
    
    addData s
        
    ' Begin a new one
    newTextFile
    
End Function

Friend Function getMyPicBox() As PictureBox
    ' Return my picture box. Used by friend objects.
    Set getMyPicBox = picBox
End Function

Public Function loadPictureIntoMemory(fileLabel As String, filePath As String, colorMode As ALPHA_ColorOptions)
    ' Load this file into memory.
    Dim s As String
    s = ""
    Dim w As Long
    Dim H As Long
    Dim x As Long, Y As Long, c As Long
    Dim p As PictureBox
    
    If usingOtherSerial Then
        Set p = otherSerialObj.getMyPicBox
    Else
        Set p = getMyPicBox
    End If
    
    ' Get the dimensions.
    p.AutoSize = True
    p.AutoRedraw = True
    
    Set p.Picture = LoadPicture(filePath)
    p.ScaleMode = ScaleModeConstants.vbPixels
    w = p.ScaleWidth
    H = p.ScaleHeight
    
    Debug.Print "Image size: ", w, H
    
    
    allocateMemory fileLabel, ALPHA_DOTS, 0, False, w, H, colorMode
    
    
    s = s & "I"         ' COMMAND CODE: Write DOTS file.
    s = s & fileLabel   ' File label to write
    s = s & makeHex(H)  ' Number of rows
    s = s & makeHex(w)  ' Number of cols
    
    ' Send the pixel data.
    For Y = 0 To H
        For x = 0 To w
            c = p.Point(x, Y)
            Select Case c
                Case vbRed: s = s & "1"
                Case vbGreen: s = s & "2"
                Case vbYellow: s = s & "3"
                Case Else: s = s & "0"
            End Select
        Next
        ' End of row
        s = s & Chr(13)
    Next
    
    Debug.Print s
    
    
    addData s


End Function



Public Function SendToSign(Optional withProgress As Boolean)
    ' Send this data to the sign!
    Sleep 200
    clearErrorRegister
    
    Sleep 100
    
    
    ' Send the data to the sign.
    ' Begin with the memory data.
    If Len(pendingMemory) > 0 Then
        sendMemory withProgress
        Sleep 100
    End If
    
    ' Now send the other data.
    sendPackets withProgress
    
    
End Function

Public Function MakeString(fileLabel As String, length As Long)
    ' Configure a string file.
    If length > 125 Then Err.Raise 1, "AlphaSignComms", "ERROR: Strings can only be 125 bytes max."
    allocateMemory fileLabel, ALPHA_STRING, length, True

End Function
    
Public Function setString(fileLabel As String, Msg As String)
    
    Dim mEnc As String
    mEnc = makeAlpha(Msg)
    Debug.Print "Setting string " & fileLabel & " to (len " & Len(mEnc) & ") ", Msg
    
    If Len(mEnc) > 125 Then mEnc = Left(mEnc, 125) ' limit to 125 characters
    
    ' Set string data.
    transBegin   ' begin the transaction.
    signOut "G"         ' Set STRING
    signOut fileLabel   ' File label
    signOut mEnc
    transEnd True
    

End Function

Public Function setRunSequence(seq As String)
    ' Set the order of the pages to run. This can be a single page or multiple pages, like "A", or "AAAAB", or "ABC".
    Sleep 200
        
        
    Debug.Print "RUN SEQUENCE: ", seq
    LastRunSequence = seq
    

    ' Set the run sequence.
    transBegin   ' begin the transaction.
    signOut "E"         ' Special function
    signOut "."         ' Set run sequence
    signOut "T"         ' S=Run in order, T=Run based on times, D=Run and delete (wonky)
    signOut "U"         ' Unlocked
    signOut seq
    
    transEnd

End Function

Public Function setRunTime(fileLabel As String, startTime As String, stopTime As String)
    ' Tell the sign if a particular page should be allowed to run or not.
    
    
    transBegin   ' begin the transaction.
    
    signOut "E"
    signOut ")"
    signOut fileLabel
    signOut startTime
    signOut stopTime
    transEnd
    

End Function

Public Function setRunDay(fileLabel As String, runYes As Boolean)
    ' Same as above but for days
    
    transBegin   ' begin the transaction.
    
    signOut "E"
    signOut "2"
    signOut fileLabel
    signOut IIf(runYes, "A", "B")
    signOut "1"
    transEnd
    

End Function

Public Function SendBeep(Optional beepCount As Long = 3)
    
    transBegin   ' begin the transaction.
    
    signOut "E"
    signOut "("
    'signOut "1" ' 3 beeps
    '        2   - tone
    '         02 - frequency (not supported on PPD)
    '           2 - duration, 0 to F
    '            2 - repeat (so 2 means beep 3 times)
    signOut "2022" & Hex(Max(0, Min(15, beepCount - 1)))
    
    transEnd
    
    
    
End Function

Public Function clearErrorRegister()
    
    
    transBegin   ' begin the transaction.
    signOut "E"
    signOut "4"
    transEnd
    

End Function



Public Function GetFreeMem()
    ' Ask for the amount of free memory. Not really needed.
    
    
    transBegin   ' begin the transaction.
    signOut "F"
    signOut "$"
    transEnd

End Function

Public Function CheckIfSignIsAlive()
    ' Ask for the speaker status to see if the sign is alive.
    
    LastLiveCheck = Now
    
    Debug.Print "Checking if sign is alive."
    transBegin   ' begin the transaction.
    signOut "F"
    signOut "!"
    transEnd

End Function


Public Function GetSerialErrorStatus()
    ' Ask for the serial error status
    
    
    transBegin   ' begin the transaction.
    signOut "F"
    signOut "*"
    transEnd

End Function

Public Function GetMemoryConfig()
    ' Not really needed.
    
    transBegin   ' begin the transaction.
    signOut "F"
    signOut "$"
    transEnd

End Function


' *****************************************************************************
' *****************************************************************************
' PRIVATE MEMBERS
' *****************************************************************************
' *****************************************************************************

Private Function sendMemory(Optional withProgress As Boolean)
    
    Debug.Print "Flushing memory queue..."
    
    ' Send the pending memory dump to the sign.
    transBegin    ' begin the transaction.
    
    signOut "E"         ' COMMAND CODE: Special function
    signOut "$"         ' Set memory.
        
    signOut pendingMemory   ' Pending memory data.

    pendingMemory = ""
    ' End of memory set up code.
    transEnd
    

End Function

Public Function setSignSerialAddress(sendToSignAddress As String, newAddress As String)
    
    ' Send the pending memory dump to the sign.
    
    
    transBegin , sendToSignAddress      ' begin the transaction.
    
    signOut "E"         ' COMMAND CODE: Special function
    signOut "7"         ' Set address
        
    signOut newAddress

    
    ' End of memory set up code.
    transEnd
    

End Function


Private Function sendPackets(Optional withProgress As Boolean)
    Dim i As Long
    
    
    ' Try the nested format.
    transBegin True       ' begin the transaction.

    Dim dataPacket As String
    Dim checkSum As String
    
    For i = 1 To UBound(pendingData)

        Debug.Print Format(Timer, "0.0000"), "Sending data packet", i, "of", UBound(pendingData), "(size " & Len(pendingData(i)) & ")"

        ' Uncomment to have progress indicators.
        If withProgress Then
            RaiseEvent Progress(((i - 1) / UBound(pendingData)) * 100)
            'frmProgress.PB1.value =
            'frmProgress.Refresh
           ' DoEvents
        End If
    
        ' Include the checksum.
        checkSum = calcCheckSum(Chr(2) & pendingData(i) & Chr(3))
        
        signOut Chr(2)  ' Start of text.
        Sleep 120       ' Wait.
    
        Debug.Print "Checksum: ", checkSum
        
        
        ' Send packet data.
        signOut pendingData(i)
        
        ' Send ETX to indicate end of text
        signOut Chr(3)
        
        ' Enable this to send the checksum
        ' Note that this doesn't appear to be working.
        
        'signOut checkSum
        
        'signOut "AAAA"
        

        ' Send ETX only if there is no next packet.
        'If i < UBound(pendingData) Then signOut Chr(3)
        
        Sleep 120
        
    Next
    
    ' Clear the pending data queue.
    ReDim pendingData(0)

    transEnd      ' End transmission.
    
    
        
    
    

End Function




Private Function makeHex(inNum As Long, Optional numDigits As Long = 2) As String
    ' Turns a number into a two-digit hex string.
    Dim s As String
    s = Hex(inNum)
    Do Until Len(s) >= numDigits
        s = "0" & s
    Loop
    
    If Len(s) > numDigits Then s = Right(s, numDigits)
    
    makeHex = s
End Function


Private Function allocateMemory(fileLabel As String, fileType As ALPHA_FileTypes, byteSize As Long, isLocked As Boolean, Optional Width As Long, Optional Height As Long, Optional colorMode As ALPHA_ColorOptions, Optional startTime As String = "FF", Optional stopTime As String = "00")

    ' Allocate memory on the Alpha sign.
    Dim s As String
    s = ""
    
    s = s & fileLabel                   ' Begin with the file label
    Select Case fileType                ' File type
        Case ALPHA_TEXT
            s = s & "A"                         ' Type: Text
            s = s & IIf(isLocked, "L", "U")     ' Locked?
            s = s & makeHex(byteSize, 4)         ' File size in hex bytes
            s = s & startTime & stopTime
        
        Case ALPHA_STRING
            s = s & "B"                         ' Type: String
            s = s & "L"                         ' Strings are always locked.
            s = s & makeHex(byteSize, 4)        ' File size in hex bytes
            s = s & "0000"                      ' Placeholder.
        
        Case ALPHA_DOTS
            s = s & "D"                         ' Type: Dots
            s = s & IIf(isLocked, "L", "U")     ' Locked?
            s = s & makeHex(Height)             ' Number of rows
            s = s & makeHex(Width)              ' Number of rows
            Select Case colorMode               ' Color mode of graphic
                Case ALPHA_3Color: s = s & "2000"
                Case ALPHA_8Color: s = s & "4000"
                Case Else: s = s & "1000"                ' monochrome
            End Select

    End Select
        
    ' Add to the memory list.
    pendingMemory = pendingMemory & s
    
    Debug.Print "Added ", Len(s), " bytes to the pending memory queue."


End Function

Private Function addData(inDat As String)
    ' Add to the pending list
    Dim n As Long
    n = UBound(pendingData) + 1
    ReDim Preserve pendingData(n)
    pendingData(n) = inDat
    
End Function


Private Function makeAlpha(inTxt As String) As String
    ' Convert codes in the text into the byte sequences that Alpha signs expect.
     
    makeAlpha = inTxt
    ' SPEED
    makeAlpha = Replace(makeAlpha, "<SPEED:1>", Chr(21))
    makeAlpha = Replace(makeAlpha, "<SPEED:2>", Chr(22))
    makeAlpha = Replace(makeAlpha, "<SPEED:3>", Chr(23))
    makeAlpha = Replace(makeAlpha, "<SPEED:4>", Chr(24))
    makeAlpha = Replace(makeAlpha, "<SPEED:5>", Chr(25))
    
    ' COLORS
    makeAlpha = Replace(makeAlpha, "<C:RED>", Chr(28) & "1")
    makeAlpha = Replace(makeAlpha, "<C:GREEN>", Chr(28) & "2")
    makeAlpha = Replace(makeAlpha, "<C:AMBER>", Chr(28) & "3")
    makeAlpha = Replace(makeAlpha, "<C:RAIN1>", Chr(28) & "9")
    makeAlpha = Replace(makeAlpha, "<C:RAIN2>", Chr(28) & "A")
    makeAlpha = Replace(makeAlpha, "<C:COLORMIX>", Chr(28) & "B")
    makeAlpha = Replace(makeAlpha, "<C:AUTO>", Chr(28) & "C")
    
    ' FONTS
    makeAlpha = Replace(makeAlpha, "<F:SANS5>", Chr(26) & "1")
    makeAlpha = Replace(makeAlpha, "<F:SANS7>", Chr(26) & "3")
    makeAlpha = Replace(makeAlpha, "<F:SERIF7>", Chr(26) & "5")
    makeAlpha = Replace(makeAlpha, "<F:SERIF16>", Chr(26) & "8")
    makeAlpha = Replace(makeAlpha, "<F:SANS16>", Chr(26) & "9")

    ' WIDE MODES
    makeAlpha = Replace(makeAlpha, "<WIDE:ON>", Chr(29) & "01")
    makeAlpha = Replace(makeAlpha, "<WIDE:OFF>", Chr(29) & "00")
    makeAlpha = Replace(makeAlpha, "<DWIDE:ON>", Chr(29) & "11")
    makeAlpha = Replace(makeAlpha, "<DWIDE:OFF>", Chr(29) & "10")
        
    ' FIXED WIDTH
    makeAlpha = Replace(makeAlpha, "<FIXEDWIDTH:ON>", Chr(29) & "41")
    makeAlpha = Replace(makeAlpha, "<FIXEDWIDTH:OFF>", Chr(29) & "40")
    
    makeAlpha = Replace(makeAlpha, "<FIXED:ON>", Chr(30) & "1")
    makeAlpha = Replace(makeAlpha, "<FIXED:OFF>", Chr(30) & "0")
    
    
       
    
    ' EFFECTS
    makeAlpha = Replace(makeAlpha, "<SCROLL>", Chr(27) & currentLine & "a")
    makeAlpha = Replace(makeAlpha, "<HOLD>", Chr(27) & currentLine & "b")
    makeAlpha = Replace(makeAlpha, "<FLASH>", Chr(27) & currentLine & "c")
    makeAlpha = Replace(makeAlpha, "<ROLL:UP>", Chr(27) & currentLine & "e")
    makeAlpha = Replace(makeAlpha, "<ROLL:DOWN>", Chr(27) & currentLine & "f")
    makeAlpha = Replace(makeAlpha, "<ROLL:LEFT>", Chr(27) & currentLine & "g")
    makeAlpha = Replace(makeAlpha, "<ROLL:RIGHT>", Chr(27) & currentLine & "h")
    makeAlpha = Replace(makeAlpha, "<ROLL:IN>", Chr(27) & currentLine & "p")
    makeAlpha = Replace(makeAlpha, "<ROLL:OUT>", Chr(27) & currentLine & "q")
    
    makeAlpha = Replace(makeAlpha, "<WIPE:UP>", Chr(27) & currentLine & "i")
    makeAlpha = Replace(makeAlpha, "<WIPE:DOWN>", Chr(27) & currentLine & "j")
    makeAlpha = Replace(makeAlpha, "<WIPE:LEFT>", Chr(27) & currentLine & "k")
    makeAlpha = Replace(makeAlpha, "<WIPE:RIGHT>", Chr(27) & currentLine & "l")
    makeAlpha = Replace(makeAlpha, "<WIPE:IN>", Chr(27) & currentLine & "r")
    makeAlpha = Replace(makeAlpha, "<WIPE:OUT>", Chr(27) & currentLine & "s")
    
    makeAlpha = Replace(makeAlpha, "<2LINESCROLLUP>", Chr(27) & currentLine & "m")      ' Rolls one line at a time up from the bottom, 2-line display mode.
    makeAlpha = Replace(makeAlpha, "<AUTO>", Chr(27) & currentLine & "o")               ' Just picks random effects
    makeAlpha = Replace(makeAlpha, "<TWINKLE>", Chr(27) & currentLine & "n" & "0")      ' The message appears, and then twinkles.
    makeAlpha = Replace(makeAlpha, "<SPARKLE>", Chr(27) & currentLine & "n" & "1")      ' Sorta like a dissolve
    makeAlpha = Replace(makeAlpha, "<SNOW>", Chr(27) & currentLine & "n" & "2")         ' Message "snows" from the top.
    makeAlpha = Replace(makeAlpha, "<INTERLOCK>", Chr(27) & currentLine & "n" & "3")    ' Pulls apart sideways and reassembles
    makeAlpha = Replace(makeAlpha, "<SWITCH>", Chr(27) & currentLine & "n" & "4")       ' Like Interlock, except top-bottom
    makeAlpha = Replace(makeAlpha, "<SLIDE>", Chr(27) & currentLine & "n" & "5")        ' Spells message from right
    makeAlpha = Replace(makeAlpha, "<SPRAY>", Chr(27) & currentLine & "n" & "6")        ' Firehoses from the right
    makeAlpha = Replace(makeAlpha, "<STARBURST>", Chr(27) & currentLine & "n" & "7")    ' Firework bangs to appear
    
    ' ANIMATIONS
    makeAlpha = Replace(makeAlpha, "<ANIM:WELCOME>", Chr(27) & currentLine & "n" & "8")
    makeAlpha = Replace(makeAlpha, "<ANIM:SLOTS>", Chr(27) & currentLine & "n" & "9")
    makeAlpha = Replace(makeAlpha, "<ANIM:THANKYOU>", Chr(27) & currentLine & "n" & "S")
    makeAlpha = Replace(makeAlpha, "<ANIM:NOSMOKING>", Chr(27) & currentLine & "n" & "U")
    makeAlpha = Replace(makeAlpha, "<ANIM:DRINKDRIVE>", Chr(27) & currentLine & "n" & "V")
    makeAlpha = Replace(makeAlpha, "<ANIM:HORSE>", Chr(27) & currentLine & "n" & "W")
    makeAlpha = Replace(makeAlpha, "<ANIM:FIREWORKS>", Chr(27) & currentLine & "n" & "X")
    makeAlpha = Replace(makeAlpha, "<ANIM:TURBOCAR>", Chr(27) & currentLine & "n" & "Y")
    makeAlpha = Replace(makeAlpha, "<ANIM:CHERRYBOMB>", Chr(27) & currentLine & "n" & "Z")
    
    ' OTHER
    makeAlpha = Replace(makeAlpha, "<DATE>", Chr(11) & "8")
    makeAlpha = Replace(makeAlpha, "<TIME>", Chr(19))
    makeAlpha = Replace(makeAlpha, "<NOHOLD>", Chr(9))
    makeAlpha = Replace(makeAlpha, "\p", Chr(12))
    makeAlpha = Replace(makeAlpha, "\n", Chr(13))
    
    
    ' Calling up other stuff
    makeAlpha = Replace(makeAlpha, "<PIC>", Chr(20))    ' Must be followed by file label
    makeAlpha = Replace(makeAlpha, "<STRING>", Chr(16)) ' Must be followed by file label

    ' Fix any messages that are too looooong.
    ' Alpha signs have issues with messages that contain words that it cant split.
    Dim i As Long
    Dim b As Long
    Dim c As Long
    Dim Msg As String
    
    For i = 1 To Len(makeAlpha)
        c = Asc(Mid(makeAlpha, i, 1))
        Select Case c
            Case Is <= 32 ' Does not count
                b = 0
            Case Else
                b = b + 1
        End Select
        
        If b > MAX_WORD_LENGTH Then
            b = 0
            Msg = Msg & " "
        End If
        
        ' Convert any special characters into the equivalent character in the Alpha font code.
        Select Case c
            Case 128: Msg = Msg & Chr(8): c = &H62  '        &#128;
            Case 129: Msg = Msg & Chr(8): c = &H48  '     ?   &#129;
            Case 130: Msg = Msg & Chr(8): c = &H48  '        &#130;
            Case 131: Msg = Msg & Chr(8): c = &H3F  '        &#131;
            Case 132: Msg = Msg & Chr(8): c = &H48  '        &#132;
            Case 133: Msg = Msg & Chr(8): c = &H48  '        &#133;
            Case 134: Msg = Msg & Chr(8): c = &H48  '        &#134;
            Case 135: Msg = Msg & Chr(8): c = &H48  '        &#135;
            Case 136: Msg = Msg & Chr(8): c = &H48  '        &#136;
            Case 137: Msg = Msg & Chr(8): c = &H48  '        &#137;
            Case 138: Msg = Msg & Chr(8): c = &H54  '        &#138;
            Case 139: Msg = Msg & Chr(8): c = &H48  '        &#139;
            Case 140: Msg = Msg & Chr(8): c = &H48  '        &#140;
            Case 141: Msg = Msg & Chr(8): c = &H48  '     ?   &#141;
            Case 142: Msg = Msg & Chr(8): c = &H56  '        &#142;
            Case 143: Msg = Msg & Chr(8): c = &H48  '     ?   &#143;
            Case 144: Msg = Msg & Chr(8): c = &H48  '     ?   &#144;
            Case 145: Msg = Msg & Chr(8): c = &H48  '        &#145;
            Case 146: Msg = Msg & Chr(8): c = &H48  '        &#146;
            Case 147: Msg = Msg & Chr(8): c = &H48  '        &#147;
            Case 148: Msg = Msg & Chr(8): c = &H48  '        &#148;
            Case 149: Msg = Msg & Chr(8): c = &H48  '        &#149;
            Case 150: Msg = Msg & Chr(8): c = &H48  '        &#150;
            Case 151: Msg = Msg & Chr(8): c = &H48  '        &#151;
            Case 152: Msg = Msg & Chr(8): c = &H48  '        &#152;
            Case 153: Msg = Msg & Chr(8): c = &H48  '        &#153;
            Case 154: Msg = Msg & Chr(8): c = &H58  '        &#154;
            Case 155: Msg = Msg & Chr(8): c = &H48  '        &#155;
            Case 156: Msg = Msg & Chr(8): c = &H48  '        &#156;
            Case 157: Msg = Msg & Chr(8): c = &H48  '     ?   &#157;
            Case 158: Msg = Msg & Chr(8): c = &H55  '        &#158;
            Case 159: Msg = Msg & Chr(8): c = &H38  '        &#159;
            Case 160: Msg = Msg & Chr(8): c = &H48  '         &#160;
            Case 161: Msg = Msg & Chr(8): c = &H4A  '        &#161;
            Case 162: Msg = Msg & Chr(8): c = &H3B  '        &#162;
            Case 163: Msg = Msg & Chr(8): c = &H3C  '        &#163;
            Case 164: Msg = Msg & Chr(8): c = &H48  '        &#164;
            Case 165: Msg = Msg & Chr(8): c = &H3D  '        &#165;
            Case 166: Msg = Msg & Chr(8): c = &H48  '        &#166;
            Case 167: Msg = Msg & Chr(8): c = &H48  '        &#167;
            Case 168: Msg = Msg & Chr(8): c = &H48  '        &#168;
            Case 169: Msg = Msg & Chr(8): c = &H48  '        &#169;
            Case 170: Msg = Msg & Chr(8): c = &H48  '        &#170;
            Case 171: Msg = Msg & Chr(8): c = &H48  '        &#171;
            Case 172: Msg = Msg & Chr(8): c = &H48  '        &#172;
            Case 173: Msg = Msg & Chr(8): c = &H48  '        &#173;
            Case 174: Msg = Msg & Chr(8): c = &H48  '        &#174;
            Case 175: Msg = Msg & Chr(8): c = &H48  '        &#175;
            Case 176: Msg = Msg & Chr(8): c = &H49  '        &#176;
            Case 177: Msg = Msg & Chr(8): c = &H48  '        &#177;
            Case 178: Msg = Msg & Chr(8): c = &H48  '        &#178;
            Case 179: Msg = Msg & Chr(8): c = &H48  '        &#179;
            Case 180: Msg = Msg & Chr(8): c = &H48  '        &#180;
            Case 181: Msg = Msg & Chr(8): c = &H48  '        &#181;
            Case 182: Msg = Msg & Chr(8): c = &H48  '        &#182;
            Case 183: Msg = Msg & Chr(8): c = &H48  '        &#183;
            Case 184: Msg = Msg & Chr(8): c = &H48  '        &#184;
            Case 185: Msg = Msg & Chr(8): c = &H48  '        &#185;
            Case 186: Msg = Msg & Chr(8): c = &H48  '        &#186;
            Case 187: Msg = Msg & Chr(8): c = &H48  '        &#187;
            Case 188: Msg = Msg & Chr(8): c = &H48  '        &#188;
            Case 189: Msg = Msg & Chr(8): c = &H48  '        &#189;
            Case 190: Msg = Msg & Chr(8): c = &H48  '        &#190;
            Case 191: Msg = Msg & Chr(8): c = &H48  '        &#191;
            Case 192: Msg = Msg & Chr(8): c = &H5B  '        &#192;
            Case 193: Msg = Msg & Chr(8): c = &H5A  '        &#193;
            Case 194: Msg = Msg & Chr(8): c = &H2F  '        &#194;
            Case 195: Msg = Msg & Chr(8): c = &H2E  '        &#195;
            Case 196: Msg = Msg & Chr(8): c = &H2E  '        &#196;
            Case 197: Msg = Msg & Chr(8): c = &H2F  '        &#197;
            Case 198: Msg = Msg & Chr(8): c = &H32  '        &#198;
            Case 199: Msg = Msg & Chr(8): c = &H20  '        &#199;
            Case 200: Msg = Msg & Chr(8): c = &H30  '        &#200;
            Case 201: Msg = Msg & Chr(8): c = &H30  '        &#201;
            Case 202: Msg = Msg & Chr(8): c = &H5E  '        &#202;
            Case 203: Msg = Msg & Chr(8): c = &H5E  '        &#203;
            Case 204: Msg = Msg & Chr(8): c = &H48  '        &#204;
            Case 205: Msg = Msg & Chr(8): c = &H5F  '        &#205;
            Case 206: Msg = Msg & Chr(8): c = &H2C  '        &#206;
            Case 207: Msg = Msg & Chr(8): c = &H2B  '        &#207;
            Case 208: Msg = Msg & Chr(8): c = &H53  '        &#208;
            Case 209: Msg = Msg & Chr(8): c = &H45  '        &#209;
            Case 210: Msg = Msg & Chr(8): c = &H35  '        &#210;
            Case 211: Msg = Msg & Chr(8): c = &H42  '        &#211;
            Case 212: Msg = Msg & Chr(8): c = &H33  '        &#212;
            Case 213: Msg = Msg & Chr(8): c = &H60  '        &#213;
            Case 214: Msg = Msg & Chr(8): c = &H39  '        &#214;
            Case 215: Msg = Msg & Chr(8): c = &H48  '        &#215;
            Case 216: Msg = Msg & Chr(8): c = &H4D  '        &#216;
            Case 217: Msg = Msg & Chr(8): c = &H37  '        &#217;
            Case 218: Msg = Msg & Chr(8): c = &H43  '        &#218;
            Case 219: Msg = Msg & Chr(8): c = &H36  '        &#219;
            Case 220: Msg = Msg & Chr(8): c = &H3A  '        &#220;
            Case 221: Msg = Msg & Chr(8): c = &H48  '        &#221;
            Case 222: Msg = Msg & Chr(8): c = &H48  '        &#222;
            Case 223: Msg = Msg & Chr(8): c = &H57  '        &#223;
            Case 224: Msg = Msg & Chr(8): c = &H25  '        &#224;
            Case 225: Msg = Msg & Chr(8): c = &H40  '        &#225;
            Case 226: Msg = Msg & Chr(8): c = &H23  '        &#226;
            Case 227: Msg = Msg & Chr(8): c = &H5D  '        &#227;
            Case 228: Msg = Msg & Chr(8): c = &H24  '        &#228;
            Case 229: Msg = Msg & Chr(8): c = &H26  '        &#229;
            Case 230: Msg = Msg & Chr(8): c = &H31  '        &#230;
            Case 231: Msg = Msg & Chr(8): c = &H27  '        &#231;
            Case 232: Msg = Msg & Chr(8): c = &H2A  '        &#232;
            Case 233: Msg = Msg & Chr(8): c = &H22  '        &#233;
            Case 234: Msg = Msg & Chr(8): c = &H28  '        &#234;
            Case 235: Msg = Msg & Chr(8): c = &H29  '        &#235;
            Case 236: Msg = Msg & Chr(8): c = &H2D  '        &#236;
            Case 237: Msg = Msg & Chr(8): c = &H41  '        &#237;
            Case 238: Msg = Msg & Chr(8): c = &H2C  '        &#238;
            Case 239: Msg = Msg & Chr(8): c = &H2B  '        &#239;
            Case 240: Msg = Msg & Chr(8): c = &H61  '        &#240;
            Case 241: Msg = Msg & Chr(8): c = &H44  '        &#241;
            Case 242: Msg = Msg & Chr(8): c = &H35  '        &#242;
            Case 243: Msg = Msg & Chr(8): c = &H42  '        &#243;
            Case 244: Msg = Msg & Chr(8): c = &H33  '        &#244;
            Case 245: Msg = Msg & Chr(8): c = &H61  '        &#245;
            Case 246: Msg = Msg & Chr(8): c = &H34  '        &#246;
            Case 247: Msg = Msg & Chr(8): c = &H48  '        &#247;
            Case 248: Msg = Msg & Chr(8): c = &H4C  '        &#248;
            Case 249: Msg = Msg & Chr(8): c = &H37  '        &#249;
            Case 250: Msg = Msg & Chr(8): c = &H43  '        &#250;
            Case 251: Msg = Msg & Chr(8): c = &H48  '        &#251;
            Case 252: Msg = Msg & Chr(8): c = &H21  '        &#252;
            Case 253: Msg = Msg & Chr(8): c = &H38  '        &#253;
            Case 254: Msg = Msg & Chr(8): c = &H48
            Case 255: Msg = Msg & Chr(8): c = &H48
        End Select
        
        Msg = Msg & Chr(c)
    Next
    
    makeAlpha = Msg
End Function

Friend Function signOut(Msg As String)
    ' Send data to serial port.
    
    outputBuffer = outputBuffer & Msg
    
    If usingOtherSerial Then
        otherSerialObj.signOut Msg
    Else
        Select Case COMM_MODE
            Case eSerial
                'On Error Resume Next
                
                If USE_COMM_BUFFER Then
                    COMM_PACKET = COMM_PACKET & Msg
                Else
                    COMM_OBJ.Output = Msg
                End If
                
                
            Case eTCP
                
                If TCP_OBJ.State <> sckConnected Then checkComms
                
                If COMMREADY Then
                    TCP_OBJ.SendData Msg
                Else
                    TCP_BUFFER = TCP_BUFFER & Msg
                End If
                
        End Select
    End If
End Function

Private Function transBegin(Optional noSTX As Boolean, Optional overwriteAddress As String)
    Dim i As Long
    
    COMM_PACKET = ""
    
    For i = 1 To 20
        signOut Chr(0)
    Next
    signOut Chr(1) ' start of header
    signOut "Z"    ' type code: all signs
    If overwriteAddress <> "" Then
        signOut overwriteAddress   ' sign address: 00 is all signs
    Else
        signOut mySignID   ' sign address: 00 is all signs
    End If
    
    outputBuffer = ""
    
    If Not noSTX Then signOut Chr(2) ' start of text

    
    

End Function

Private Function transEnd(Optional sendChecksum As Boolean)
    If sendChecksum Then
        'outputBuffer = outputBuffer & Chr(3)
        'Debug.Print outputBuffer, calcCheckSum(outputBuffer)
        'signOut Chr(3)
        'signOut calcCheckSum(outputBuffer)
        'outputBuffer = ""
        
    End If
    outputBuffer = ""
    signOut Chr(4)  ' End of transmission
    
    If USE_COMM_BUFFER Then
        COMM_BUFFER.Add COMM_PACKET
        COMM_PACKET = ""
        MyTimer.Enabled = True
    End If
    
    
    
End Function

Private Function processMessage(inMsg As String)
    ' Process the received message.
   On Error GoTo processMessage_Error

    Debug.Print "Msg: ", inMsg
    
    ' THIS FUNCTION IS NOT COMPLETE
    ' THIS LIBRARY CURRENTLY DOES NOT IMPLEMENT RECEIVING COMMANDS FROM THE ALPHA SIGN
    
    '     1234567890
    'Msg: 000E*@00B4
    ' Parse the message.  Start by finding the START OF TEXT
    
    Dim typeCode As String, signAddr As String
    Dim cmdCode As String
    Dim e As Long
    Dim inData As String
    Dim sfRCode As String
    Dim checkSum As String
    Dim myCheckSum As String
    Dim memTable As String
    
    Dim rr As Long
    
    Dim fileType As String
    
    
    typeCode = Mid(inMsg, 1, 1)
    signAddr = Mid(inMsg, 2, 2)
    cmdCode = Mid(inMsg, 5, 1)
    e = InStr(1, inMsg, Chr(3)) ' Find end of text
    If e > 0 Then
        inData = Mid(inMsg, 4, e - 3)
        'Debug.Print "indata!: ", inData 'E*@
        
        checkSum = Right(inMsg, Len(inMsg) - e)
        Debug.Print "Checksum: >" & checkSum & "<"
        
        myCheckSum = calcCheckSum(inData)
        Debug.Print "my checksum: " & myCheckSum
        
        
        Select Case cmdCode
            Case "E" ' Special functions return
            
                sfRCode = Mid(inData, 3, 1)
                Select Case sfRCode
                    Case "*" ' Serial error register.
                        
                        rr = Asc(Mid(inData, 4, 1))
                        
                        Debug.Print "Register: ", rr
                        If (rr And 1) Then Debug.Print "SERIAL ERROR: Parity error (not even parity)"
                        If (rr And 2) Then Debug.Print "SERIAL ERROR: Bit framing error (incorrect baud rate)"
                        If (rr And 4) Then Debug.Print "SERIAL ERROR: Serial timeout (timeout period exceeded)"
                        If (rr And 8) Then Debug.Print "SERIAL ERROR: Insufficient serial buffer space"
                        If (rr And 16) Then Debug.Print "SERIAL ERROR: Serial checksum error"
                        If (rr And 32) Then Debug.Print "SERIAL ERROR: Illegal command code, file label, illegal read or write SPECIAL FUNCTION command"
                        If (rr And 64) = 0 Then Debug.Print "SERIAL: Code 1 is 0!"
                        If (rr And 128) Then Debug.Print "SERIAL ERROR: Code is 0 is 1!"
                        
                    Case "$" ' Memory configuration
                        'E$!AU002CFF00aBL00320000bBL00320000cBL00320000eBL001E0000dBL00320000fBL00320000gBL007D00001BL007D00002BL007D00003BL007D00004BL007D0000AAU0033FF00BAU0031FF00CAU0030FF00DAU0036FF00EAU002BFF00FAU0022FF00GAU00D8FF00@AU000CFF00
                        '    L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X L---------X
                        'E$ !AU002CFF00 aBL00320000 bBL00320000 cBL00320000 eBL001E0000 dBL00320000 fBL00320000 gBL007D0000 1BL007D0000 2BL007D0000 3BL007D0000 4BL007D0000 AAU0033FF00 BAU0031FF00 CAU0030FF00 DAU0036FF00 EAU002BFF00 FAU0022FF00 GAU00D8FF00 @AU000CFF00
                        
                        inData = Right(inData, Len(inData) - 3)
                        Do While Len(inData) >= 11
                            memTable = Left(inData, 11)
                            
                            ' Parse this message.
                            
                            Debug.Print "Label: " & Left(memTable, 1);
                            Select Case Mid(memTable, 2, 1) ' Type
                                Case "A": Debug.Print " Type: TEXT    ";
                                Case "B": Debug.Print " Type: STRING  ";
                                Case "C": Debug.Print " Type: DOTS    ";
                            End Select
                            
                            Select Case Mid(memTable, 3, 1) ' Protected
                                Case "U": Debug.Print " [Unlocked] ";
                                Case "L": Debug.Print " [Locked]   ";
                            End Select
                            
                            Debug.Print "Size: " & Val("&H" & Mid(memTable, 4, 4));
                            
                                        
                            Debug.Print ""
                            
                            inData = Right(inData, Len(inData) - 11)
                            
                        
                        Loop
                    
                    Case "!"
                        Debug.Print "SIGN IS PRESENT"
                        LastLiveResponse = Now

                End Select
                
                
            
        
        End Select
        
    End If
    
    
    

   On Error GoTo 0
   Exit Function

processMessage_Error:

    Debug.Print "[ERROR]", "In processMessage of Class Module AlphaSignComms", Err.Number, Err.Description


End Function

Private Function calcCheckSum(inStrng As String) As String

    ' Create the checksum.
    Dim n As Long
    Dim i As Long
    For i = 1 To Len(inStrng)
        n = n + Asc(Mid(inStrng, i, 1))
    Next
    
    ' Split into two bytes.
    calcCheckSum = Format(Hex(n \ 256), "00") & Format(Hex(n Mod 256), "00")
    
    
    

End Function
    

Private Sub COMM_OBJ_OnComm()

    Dim a As String
    Select Case COMM_OBJ.CommEvent
        Case comEvReceive
        
            ' print received data
            COMM_OBJ.InputLen = 0
            a = COMM_OBJ.Input
            
            Receive a
    
    End Select

End Sub

Private Sub MyTimer_Timer()
    
    Dim Msg As String
    If COMM_BUFFER.Count > 0 Then
        If Not IsTransmitting Then
            Msg = COMM_BUFFER(1)
            COMM_BUFFER.Remove 1
            
            IsTransmitting = True
            On Error Resume Next
            
            COMM_OBJ.Output = Msg
            'Put #11, , Msg

        
        
        End If
    Else
        MyTimer.Enabled = False
    End If
    
End Sub

Private Sub TCP_OBJ_Connect()
    Debug.Print "!! TCP CONNECTED!"
    COMMREADY = True
    If TCP_BUFFER <> "" Then
        Debug.Print "CLEARING BUFFER!"
        signOut TCP_BUFFER
        TCP_BUFFER = ""
    End If
    
    If Not firstTCP Then
        firstTCP = True
        RaiseEvent TcpConnectionActive
    End If
End Sub

Private Sub TCP_OBJ_DataArrival(ByVal bytesTotal As Long)
    Dim a As String
    TCP_OBJ.GetData a
    Receive a
    
End Sub

Private Sub checkComms()
    If TCP_OBJ.State <> sckConnected Then
        COMMREADY = False
        If Now - lastConnAttempt > 10 * OneSecond Then
            Debug.Print "No connection. Attempting to reconnect..."
            openComm
        End If
    End If
End Sub

Public Function isSignConnected() As Boolean
    
    If Not TCP_OBJ Is Nothing Then
        isSignConnected = (TCP_OBJ.State = sckConnected)
        If Not isSignConnected Then
            checkComms
        End If
    Else
        isSignConnected = True
    End If
End Function
    

Private Function safeRight(inSt As String, numChr As Long) As String
    If Len(inSt) >= numChr And numChr > 0 Then
        safeRight = Right(inSt, numChr)
    Else
        safeRight = inSt
    End If
End Function

Private Function Max(n1, n2)
    If n1 > n2 Then
        Max = n1
    Else
        Max = n2
    End If
End Function

Private Function Min(n1, n2)
    If n1 < n2 Then
        Min = n1
    Else
        Min = n2
    End If
End Function

